#!/usr/bin/python
#
# Author: Matthias Gerstner
#
# All of this source code is licensed under:
#
# ISC License
#
# Copyright (c) 2021, SUSE LLC
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import errno
import socket
import ssl
import time

import barrier_common as common
from barrier_common import ClipFormat, InMessage, OutMessage, TransferStatus,\
    fprint


class BarrierClient(common.BarrierBase):

    def __init__(self):
        # operate in the oldest version by default
        super().__init__(minor_version=0)
        self.m_parser.add_argument(
            "remote", default="vm-tw", nargs='?'
        )
        self.m_parser.add_argument(
            "--client-name", default="funnyclient",
            help="Client name to report to server in Hello message"
        )
        self.m_parser.add_argument(
            "--screen-info",
            default="0,0,1024,768,0,0",
            metavar='X,Y,W,H,MX,MY',
            help="Screen information to report in the QINF->DINF message reply"
        )
        self.m_parser.add_argument(
            "--honor-keepalive",
            default=False,
            action='store_true',
            help="If set then keepalive message requests from the server will be answered."
        )
        self.m_parser.add_argument(
            "--send-file",
            metavar='PATH',
            default=None,
            help="Send the given file to the server"
        )
        self.m_parser.add_argument(
            "--send-clipboard",
            metavar='STRING',
            default=None,
            help="Send the given string as clipboard data"
        )

        self.m_parser.add_argument("--client-cert")
        self.m_parser.add_argument("--client-key")

        self.m_parser.add_argument(
            "--run-hello-loop",
            action='store_true',
            help="Quickly open connections and receive/send hello, closing immediately again"
        )
        self.m_parser.add_argument(
            "--send-infinite-message",
            action='store_true',
            help="Send a message of the maximum message length containing garbage data"
        )
        self.m_parser.add_argument(
            "--run-open-close-loop",
            action='store_true',
            help="Open and close a lot of connections without transmitting any data"
        )
        self.m_parser.add_argument(
            "--trigger-timer-race",
            action='store_true',
            help="Attempts to trigger a server race condition waiting for the QINF reply, when the 30 second timer hits in parallel with the reply. Can take a long time to trigger."
        )

    def processArgs(self):
        super().processArgs()

        if bool(self.m_args.client_cert) != bool(self.m_args.client_key):
            raise Exception("need both --client-cert and --client-key")

        info = self.m_args.screen_info.split(',')
        if len(info) != 6:
            raise Exception("Expected six-element comma separated list for --screen-info")
        try:
            self.m_screen_info = [int(v) for v in info]
        except ValueError:
            raise Exception("Expected integers for --screen-info")

        if self.m_args.minor_version < 5 and self.m_args.send_file:
            raise Exception("--send-file only works starting with protocol version 1.5 (see --minor-version)")

        if self.m_args.minor_version < 6 and self.m_args.send_clipboard:
            raise Exception("--send-file only works starting with protocol version 1.5 (see --minor-version)")

    def run(self, args):
        self.m_args = self.m_parser.parse_args(args)

        self.processArgs()

        if self.m_args.run_hello_loop:
            fprint("Running connection loop with Hello exchange")
            i = 0
            while True:
                self.m_sock = self.connect()
                self.exchangeHello()
                self.m_sock.close()
                i += 1
                if i % 10 == 0:
                    fprint("Connection nr.", i)
        elif self.m_args.send_infinite_message:
            self.sendInfinitely()
        elif self.m_args.run_open_close_loop:
            i = 0
            while True:
                try:
                    self.m_sock = self.connect()
                except socket.timeout:
                    fprint("SSL handshake timed out, retrying")
                    continue
                self.m_sock.close()
                i += 1
                if i % 100 == 0:
                    fprint("Connection nr.", i)
        elif self.m_args.trigger_timer_race:
            i = 0
            while True:
                try:
                    if i != 0:
                        print("\n\n")
                    i += 1
                    print(f"Attempt {i}".center(80))
                    print("=" * 80)
                    self.m_sock = self.connect()
                    print("Exchanging hello handshake")
                    fprint("Waiting ~30 seconds with reply, racing with server side timeout")
                    self.exchangeHello(reply_delay=29.90)
                except socket.timeout:
                    print("SSL handshake timed out ?")
                except OSError as e:
                    if e.errno == errno.EHOSTUNREACH:
                        fprint(f"Remote host unreachable ({e} -> firewall?)")
                    if e.errno in (errno.ECONNREFUSED, errno.ECONNRESET):
                        fprint(f"Remote host unreachable ({e} -> crashed?)")
                    else:
                        fprint("OSError occured:", e)
                except Exception as e:
                    print("Connection failed:", type(e), e)
        else:
            self.m_sock = self.connect()
            self.readLoop()

        self.m_sock.close()

    def sessionReady(self):

        self.sendGrabClipboard()

        if self.m_args.send_file:
            with open(self.m_args.send_file, 'rb') as fd:
                data = fd.read()

            self.sendFile(data)
        if self.m_args.send_clipboard:
            self.sendClipboard(0, 50, self.m_args.send_clipboard)

    def readLoop(self):
        msg = InMessage()

        self.exchangeHello()
        #self.sendInfo()

        while True:
            msg.readMessage(self.m_sock)
            code = msg.getCode()
            if code == "DSOP":
                self.printOptions(msg)
            elif code == "CROP":
                fprint("Received options reset request")
            elif code == "CCLP":
                self.printClipboardGrab(msg)
            elif code == "CINN":
                self.printEnterMessage(msg)
            elif code == "DMMV":
                self.printMouseMoveMessage(msg)
            elif code in ("DKUP", "DKDN"):
                self.printKeyMessage(msg)
            elif code == "CALV":
                fprint("Received keepalive request")
                if self.m_args.honor_keepalive:
                    self.sendKeepalive()
            elif code == "QINF":
                self.sendInfo()
                self.sessionReady()
            else:
                fprint(f"Received '{code}' message")

    def printKeyMessage(self, msg):

        key = msg.parseUint16()
        mask = msg.parseUint16()
        button = msg.parseUint16()
        event = "down" if msg.getCode() == "DKDN" else "up"
        char = chr(key)
        if not char.isalnum():
            char = hex(key)

        print(f"Received key {event}: key = {key}, mask = {mask}, button = {button}, character = {char}")

    def printOptions(self, msg):
        # this is pretty strange, pairs of 32-bit integers are sent, the keys
        # are actually four-character-sequences converted to 32-bit integers.

        length = msg.parseUint32()

        if msg.remainingBytes() / 4 != length:
            print(f"Received options with invalid length. Says {length} items but only {msg.remainingBytes()} bytes present")
            return

        if length % 2 != 0:
            print("Received uneven amount of integers in options list")
            return

        print(f"Received options comprised of {int(length/2)} pairs:\n")

        while msg.remainingBytes() != 0:
            key = msg.getNextBytes(4).decode('ascii')
            val = msg.parseUint32()

            print(key, val)

    def printEnterMessage(self, msg):
        x_abs = msg.parseInt16()
        y_abs = msg.parseInt16()
        seq = msg.parseUint32()
        mask = msg.parseUint16()

        print(f"Received enter message: x = {x_abs}, y = {y_abs}, seq = {seq}, mask = {mask}")

    def printMouseMoveMessage(self, msg):
        x_abs = msg.parseInt16()
        y_abs = msg.parseInt16()

        print(f"Received mouse move message: x = {x_abs}, y = {y_abs}")

    def connect(self):
        context = ssl.SSLContext()

        if self.m_args.client_cert:
            context.load_cert_chain(
                self.m_args.client_cert,
                keyfile=self.m_args.client_key
            )

        conndetails = (self.m_args.remote, self.m_args.port)

        with socket.create_connection(conndetails) as sock:
            # the SSL handshake stalls roughly every second attempt, not sure
            # why that is, it is probably in the internals of the daemon's SSL
            # impl.
            sock.settimeout(2.5)
            ret = context.wrap_socket(
                sock,
                server_hostname=self.m_args.remote
            )
            # disable the timeout again once the handshake is through
            ret.settimeout(None)
            return ret

    def _handleDelay(self, reply_delay):
        if reply_delay is None:
            return

        time.sleep(reply_delay)

    def exchangeHello(self, reply_delay=None):
        code = "Barrier"
        msg = self.readMessage(code)
        major, minor = msg.parseInt16(), msg.parseInt16()
        fprint(f"Remote is barrier {major}.{minor}")

        msg = OutMessage(code)
        msg.addUint16(self.m_args.major_version)
        msg.addUint16(self.m_args.minor_version)
        msg.addString(self.m_args.client_name)

        self._handleDelay(reply_delay)

        self.sendMessage(msg)

    def sendInfo(self, reply_delay=None, wait_for_ack=True):

        reply = OutMessage("DINF")
        # there is a dummy between x,y,w,h and mx, my, so insert it
        for i in self.m_screen_info[:4] + [0] + self.m_screen_info[4:6]:
            reply.addInt16(i)

        self._handleDelay(reply_delay)

        self.sendMessage(reply)

        if wait_for_ack:
            self.waitForACK()

    def sendGrabClipboard(self):
        grab_req = OutMessage("CCLP")
        # clipboard ID (0 or 1)
        grab_req.addUint8(1)
        # sequence number, increasing according to some strategy
        grab_req.addUint32(10)

        self.sendMessage(grab_req)

    def sendClipboard(self, clip_id, sequence, data):
        def addInfo(msg, mark):
            msg.addUint8(clip_id)
            msg.addUint32(sequence)
            msg.addUint8(mark)

        # the same protocol as in sendFile(), only with an additional clip_id
        # and sequence
        # update: actually the payload data is interpreted very specially in
        # the clipboard case. It has the following structure:
        # <uint32 number of formats>[<uint32 format ClipFormat enum><uint32 #num of bytes><actual data ...>],...

        # send one format "Text"
        extra_data = 4 + 4 + 4

        msg = OutMessage("DCLP")
        addInfo(msg, TransferStatus.Start)
        msg.addString(str(len(data) + extra_data))
        self.sendMessage(msg)

        msg.dropData()
        addInfo(msg, TransferStatus.Chunk)

        inner_header = msg.getUint32(1) + msg.getUint32(ClipFormat.Text.value) + msg.getUint32(len(data))
        msg.addString(inner_header + data.encode())
        self.sendMessage(msg)

        msg.dropData()
        addInfo(msg, TransferStatus.End)
        msg.addString("")
        self.sendMessage(msg)

    def sendFile(self, data):
        # the start message contains just the file size as a string
        msg = OutMessage("DFTR")
        msg.addUint8(TransferStatus.Start)
        msg.addString(str(len(data)))
        self.sendMessage(msg)

        msg.dropData()
        msg.addUint8(TransferStatus.Chunk)
        msg.addString(data)
        self.sendMessage(msg)

        msg.dropData()
        msg.addUint8(TransferStatus.End)
        msg.addString("")
        self.sendMessage(msg)

    def printClipboardGrab(self, msg):
        if msg.remainingBytes() != 5:
            print(f"Received invalid clipboard grab message of {msg.remainingBytes()} bytes")
            return

        clip_id = msg.parseUint8()
        seq_id = msg.parseUint32()

        print(f"Got clipboard grab message for id = {clip_id} and seqnr = {seq_id}")

    def sendInfinitely(self):
        self.m_sock = self.connect()
        fprint("Sending garbage data of maximum length")
        header = OutMessage.getMessageLengthHeader(OutMessage.MAXLEN)
        self.m_sock.sendall(header)
        i = 0
        chunk = b"some fun stuff" * 100
        while True:
            self.m_sock.sendall(chunk)
            i += 1

            if i % 10000 == 0:
                amount = common.humanReadableBytes(len(chunk) * i)
                fprint(f"Sent {amount} of data")


if __name__ == "__main__":
    common.runMain(BarrierClient)
