# -*- coding: utf-8 -*-
# vim: ts=4 et sw=4 sts=4 :

# Author: Matthias Gerstner
#
# All of this source code is licensed under:
#
# ISC License
#
# Copyright (c) 2021, SUSE LLC
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import argparse
import struct
import sys
from enum import IntEnum


class ConnectionClosed(Exception):
    def __init__(self, text):
        super().__init__(text)


TransferStatus = IntEnum('TransferStatus', ['Start', 'Chunk', 'End'], start=1)
ClipFormat = IntEnum('ClipFormat', ['Text', 'HTML', 'Bitmap'], start=0)
DEFPORT = 24800


def fprint(*args, **kwargs):
    print(*args, **kwargs)
    sys.stdout.flush()


def eprint(*args, **kwargs):
    kwargs["file"] = sys.stderr
    print(*args, **kwargs)


def printEx(e):
    import traceback
    _, _, tb = sys.exc_info()

    width = None

    for t in traceback.extract_tb(tb):
        fname, lnum, fnc, line = t
        if width is None:
            width = len(fname) + 5
        else:
            print(file=sys.stderr)
        print(f"{fname.rjust(width)}:{str(lnum).rjust(4)}: {fnc}", file=sys.stderr, end='')

    print(f": {str(e)}", file=sys.stderr)


def humanReadableBytes(num):
    KILO = 1024
    MEGA = KILO * 1024
    GIGA = MEGA * 1024

    if num < KILO:
        return f"{num} bytes"
    elif num < MEGA:
        num = num / KILO
        return "{:.2f} kilobytes".format(num)
    elif num < GIGA:
        num = num / MEGA
        return "{:.2f} megabytes".format(num)
    else:
        num = num / GIGA
        return "{:.2f} gigabytes".format(num)


verbose = False


def setVerbose(val):
    global verbose
    verbose = val


def isVerbose():
    return verbose


class OutMessage:

    MAXLEN = 2**32 - 1

    def __init__(self, code=None):
        self.clear()

        if code:
            self.addCode(code)

    def clear(self):
        self.m_msg = bytes()
        self.m_code = ""

    def dropData(self):
        code = self.m_code
        self.clear()
        self.addCode(code)

    @classmethod
    def getUint32(cls, num):
        return struct.pack(">I", num)

    def addUint32(self, num):
        self.m_msg += self.getUint32(num)

    def prependUint32(self, num):
        self.m_msg = self.getUint32(num) + self.m_msg

    def addUint16(self, num):
        self.m_msg += struct.pack(">H", num)

    def addInt16(self, num):
        self.m_msg += struct.pack(">h", num)

    def addUint8(self, num):
        self.m_msg += bytes([num])

    @classmethod
    def getMessageLengthHeader(cls, ln):
        return cls.getUint32(ln)

    def addCode(self, s):
        self.m_msg += s.encode()
        self.m_code = s

    def getCode(self):
        return self.m_code

    def addString(self, s):
        self.addUint32(len(s))
        if isinstance(s, str):
            s = s.encode()
        self.m_msg += s

    def getRaw(self):
        return self.getMessageLengthHeader(len(self.m_msg)) + self.m_msg

    def sendMessage(self, sock):
        raw = self.getRaw()
        if isVerbose():
            fprint(f"> Sending '{self.getCode()}' message of {len(raw)} bytes length.")
        sock.sendall(raw)


class InMessage:

    HEADER_LEN = 4

    def __init__(self):
        self.reset()

    def reset(self, code=""):
        self.m_msg = bytes()
        self.m_msg_len = 0
        self.m_pos = 0
        self.m_code = code

    def getCode(self):
        return self.m_code

    def parseUint32(self):
        data = self.getNextBytes(4)
        return struct.unpack(">I", data)[0]

    def parseInt16(self):
        data = self.getNextBytes(2)
        return struct.unpack(">h", data)[0]

    def parseUint16(self):
        data = self.getNextBytes(2)
        return struct.unpack(">H", data)[0]

    def parseUint8(self):
        data = self.getNextBytes(1)
        return data[0]

    def parseString(self):
        length = self.parseUint32()
        s = self.getNextBytes(length)
        return s.decode()

    def length(self):
        return len(self.m_msg)

    def payloadLength(self):
        # subtract message length header and
        return len(self.m_msg) - 4 - len(self.m_code)

    def remainingBytes(self):
        return self.length() - self.m_pos

    def getNextBytes(self, num):
        assert self.remainingBytes() >= num
        data = self.m_msg[self.m_pos:self.m_pos + num]
        self.m_pos += num
        return data

    def _readMessageLength(self, sock):
        while self.length() != self.HEADER_LEN:
            chunk = sock.recv(self.HEADER_LEN - self.length())
            if chunk is None or len(chunk) == 0:
                raise ConnectionClosed("Remote closed connection")
            self.m_msg += chunk

        self.m_msg_len = self.parseUint32()

        if self.m_msg_len > 1024 * 64:
            raise Exception(f"Excess message length {self.m_msg_len} received")

    def guessCode(self):
        code = ""

        pos = self.m_pos

        # except from the initial "Barrier" message all other message codes
        # seem to be four bytes long
        while pos < len(self.m_msg) and len(code) < 4:
            byte = self.m_msg[pos]
            char = chr(byte)
            if char.isalnum():
                code += char
                pos += 1
            else:
                break

        self.m_pos += len(code)
        self.m_code = code

    def readMessage(self, sock, code=""):
        self.reset(code)

        self._readMessageLength(sock)

        while self.remainingBytes() < self.m_msg_len:
            self.m_msg += sock.recv(self.m_msg_len - self.remainingBytes())

        if self.remainingBytes() < len(code):
            raise Exception(f"Received only {self.remainingBytes()} trying to read {code} message: {self.m_msg.decode()}")

        if code:
            cmp_code = self.getNextBytes(len(code))

            if cmp_code.decode() != code:
                raise Exception(f"Received unexpected message code {cmp_code} instead of {code}")
        else:
            self.guessCode()

        if isVerbose():
            fprint(f"< Received '{self.m_code}' message with {self.payloadLength()} payload bytes.")


class BarrierBase:

    def __init__(self, minor_version):
        self.m_parser = argparse.ArgumentParser()
        self.m_parser.add_argument(
            "-p", "--port", default=DEFPORT, type=int
        )
        self.m_parser.add_argument(
            "-v", "--verbose", default=False, action='store_true'
        )
        self.m_parser.add_argument(
            "--minor-version", default=minor_version, type=int,
            help="Minor version number to report to remote party"
        )
        self.m_parser.add_argument(
            "--major-version", default=1, type=int,
            help="Major version number to report to remote party"
        )

    def processArgs(self):
        setVerbose(self.m_args.verbose)

    def readMessage(self, code=""):
        msg = InMessage()
        msg.readMessage(self.m_sock, code)

        return msg

    def sendMessage(self, msg):
        msg.sendMessage(self.m_sock)

    def waitForACK(self):
        self.readMessage("CIAK")

    def sendACK(self):
        msg = OutMessage("CIAK")
        self.sendMessage(msg)

    def sendKeepalive(self):
        msg = OutMessage("CALV")
        self.sendMessage(msg)


def runMain(clazz):

    try:
        barrier = clazz()
        barrier.run(sys.argv[1:])
    except ConnectionClosed as e:
        print(e)
    except Exception as e:
        printEx(e)
        sys.exit(1)
