#!/usr/bin/python
#
# Author: Matthias Gerstner
#
# All of this source code is licensed under:
#
# ISC License
#
# Copyright (c) 2021, SUSE LLC
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import random
import socket
import ssl
import sys

import barrier_common as common
from barrier_common import OutMessage, fprint


class BarrierServer(common.BarrierBase):

    def __init__(self):
        # the real client refuses connections to older protocol version
        super().__init__(minor_version=6)
        self.m_parser.add_argument(
            "-l", "--listen",
            required=True,
            help="On which IP address / hostname to listen for client connections.",
        )
        # these are required on the server end, otherwise we get a not very
        # helpful "no shared cipher" error message when a client attempts to
        # connect
        self.m_parser.add_argument("--server-cert", default='cert.pem')
        self.m_parser.add_argument("--server-key", default='key.pem')
        self.m_parser.add_argument("--no-ssl", help="Offer plaintext connections instead of SSL connections", action='store_true', default=False)

    def processArgs(self):
        super().processArgs()

        if bool(self.m_args.server_cert) != bool(self.m_args.server_key):
            raise Exception("need both --server-cert and --server-key")

    def run(self, args):

        self.m_args = self.m_parser.parse_args(args)
        self.processArgs()
        self.setupSocket()
        while True:
            self.acceptClient()

    def setupSocket(self):
        #self.m_ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        self.m_ssl_context = ssl.SSLContext()

        if self.m_args.server_cert:
            print("Loading certificate and private key from",
                  self.m_args.server_cert, "and", self.m_args.server_key)
            self.m_ssl_context.load_cert_chain(
                self.m_args.server_cert,
                keyfile=self.m_args.server_key
            )
        elif not self.m_args.no_ssl:
            print("Cannot run in SSL mode without a --server-cert")
            sys.exit(1)

        self.m_listen_sock = socket.socket()
        addr, port = self.m_args.listen, self.m_args.port
        self.m_listen_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.m_listen_sock.bind((addr, port))
        self.m_listen_sock.listen()
        print(f"Listening for client connections on {addr}:{port}")
        fprint()

    def acceptClient(self):

        sock, peer = self.m_listen_sock.accept()
        fprint(f"New connection from {peer}")
        if self.m_args.no_ssl:
            self.m_sock = sock
        else:
            try:
                sock.settimeout(2.5)
                #print(self.m_ssl_context.get_ciphers())
                self.m_sock = self.m_ssl_context.wrap_socket(sock, server_side=True)
                self.m_sock.settimeout(None)
            except socket.timeout as e:
                fprint("SSL handshake timed out:", e)
                return

        try:
            fprint("Starting client session")
            self.handleClient()
            fprint("Ending client session")
        except common.ConnectionClosed as e:
            print(e)
            print('-' * 72)
        except Exception as e:
            fprint("Exception occured while processing client:")
            common.printEx(e)
        finally:
            try:
                self.m_sock.shutdown(socket.SHUT_RDWR)
                self.m_sock.close()
            except Exception as e:
                fprint("Failed to close connection:", e)
            del self.m_sock

    def handleClient(self):

        self.exchangeHello()
        self.sendOptions()
        self.sendEnterScreen()
        self.queryInfo()

        #self.sendKeyEvent('X')

        while True:
            msg = self.readMessage()
            if msg.getCode() == "DINF":
                screen_info = []
                for i in range(7):
                    screen_info.append(msg.parseInt16())
                fprint("Client screen info", screen_info)
            else:
                fprint(f"Received message {msg.getCode()}")

    def exchangeHello(self):
        code = "Barrier"
        msg = OutMessage(code)
        msg.addUint16(self.m_args.major_version)
        msg.addUint16(self.m_args.minor_version)

        self.sendMessage(msg)
        msg = self.readMessage(code)
        major, minor = msg.parseInt16(), msg.parseInt16()
        client = msg.parseString()

        fprint(f"Remote is barrier {major}.{minor} named '{client}'")

    def queryInfo(self):
        msg = OutMessage("QINF")
        self.sendMessage(msg)

    def sendOptions(self):
        msg = OutMessage("DSOP")

        num_opts = 0

        msg.addUint32(num_opts)

        for i in range(num_opts):
            print("Adding option")
            key = random.randint(0, (2**32 - 1))
            val = random.randint(0, (2**32 - 1))
            msg.addUint32(key)
            msg.addUint32(val)

        self.sendMessage(msg)

    def sendEnterScreen(self):
        msg = OutMessage("CINN")
        self.sendMessage(msg)

    def sendKeyEvent(self, key):
        self.sendKeyDown(key)
        self.sendKeyUp(key)

    def sendKeyDown(self, key):
        self.sendKeyUpDown(key, code="DKDN")

    def sendKeyUp(self, key):
        self.sendKeyUpDown(key, code="DKUP")

    def sendKeyUpDown(self, key, code):
        msg = OutMessage(code)
        msg.addUint16(ord(key))
        msg.addUint16(0)
        msg.addUint16(0)
        self.sendMessage(msg)


if __name__ == "__main__":
    common.runMain(BarrierServer)
